from fast_bitrix24.utils import http_build_query


class TestHttpBuildQuery:
    def test_original(self):
        assert http_build_query({"alpha": "bravo"}) == "alpha=bravo&"

        test = http_build_query({"charlie": ["delta", "echo", "foxtrot"]})
        assert "charlie[0]=delta" in test
        assert "charlie[1]=echo" in test
        assert "charlie[2]=foxtrot" in test

        test = http_build_query(
            {
                "golf": [
                    "hotel",
                    {"india": "juliet", "kilo": ["lima", "mike"]},
                    "november",
                    "oscar",
                ]
            }
        )
        assert "golf[0]=hotel" in test
        assert "golf[1][india]=juliet" in test
        assert "golf[1][kilo][0]=lima" in test
        assert "golf[1][kilo][1]=mike" in test
        assert "golf[2]=november" in test
        assert "golf[3]=oscar" in test

    def test_new(self):
        d = {"FILTER": {"STATUS_ID": "CLOSED"}}

        test = http_build_query(d)
        assert test == "FILTER[STATUS_ID]=CLOSED&"

        d = {"FILTER": {"!STATUS_ID": "CLOSED"}}

        test = http_build_query(d)
        assert test == "FILTER[%21STATUS_ID]=CLOSED&"



import asyncio
import sys

import aiohttp.client_exceptions as exc
import pytest

from fast_bitrix24 import Bitrix


@pytest.mark.skipif(sys.version_info < (3, 8), reason="requires python3.8 or higher")
@pytest.mark.parametrize("exception", [exc.ServerConnectionError, asyncio.TimeoutError])
@pytest.mark.asyncio
async def test_retries(exception):
    from unittest.mock import AsyncMock

    bitrix = Bitrix("https://google.com/path")
    srh = bitrix.srh

    srh.request_attempt = AsyncMock(spec=srh.request_attempt, side_effect=exception)

    # должна исчерпать все попытки и выдать RuntimeError
    with pytest.raises(RuntimeError):
        await srh.single_request("abc")



import logging

import pytest
from beartype.typing import Dict, List, Union
from fast_bitrix24.logger import logger
from fast_bitrix24.server_response import (
    ErrorInServerResponseException,
    ServerResponseParser,
)
from fast_bitrix24.srh import ServerRequestHandler

logger.addHandler(logging.StreamHandler())


class MockSRH(ServerRequestHandler):
    def __init__(self, response: Union[Dict, List[Dict]]):
        self.response = response if isinstance(response, list) else [response]
        self.element_no = -1

        super().__init__("https://google.com/path", None, False, 50, 2, 480, None)

    async def single_request(self, *args, **kwargs):
        self.element_no += 1
        return self.response[self.element_no]


def test_single_success():
    from tests.real_responses.crm_lead_fields_success import response

    results = ServerResponseParser(response).extract_results()
    assert isinstance(results, dict)

    from tests.real_responses.crm_lead_list_one_page_success import response

    results = ServerResponseParser(response).extract_results()
    assert isinstance(results, list)


def test_batch_success(bx_dummy, recwarn):
    from tests.real_responses.crm_lead_list_several_pages_success import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_all("crm.lead.list")
    assert isinstance(results, list)
    print(w.message for w in recwarn.list)
    with pytest.raises(Exception):
        recwarn.pop(RuntimeWarning)


def test_batch_single_page_error(bx_dummy):
    from tests.real_responses.crm_get_batch_mix_success_and_errors import response

    bx_dummy.srh = MockSRH(response)
    with pytest.raises(ErrorInServerResponseException):
        bx_dummy.get_by_ID("crm.lead.get", [0, 1, 35943])


def test_call_single_success(bx_dummy):
    from tests.real_responses.call_single_success import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.call("crm.lead.get", {"ID": 35943})
    assert isinstance(results, dict)


def test_call_several_success(bx_dummy):
    from tests.real_responses.call_several_success import response

    bx_dummy.srh = MockSRH(response)

    ID_list = [161, 35943, 171]
    results = bx_dummy.call("crm.lead.get", [{"ID": ID} for ID in ID_list])

    assert isinstance(results, tuple)
    assert len(results) == 3
    assert isinstance(results[0], dict)

    # нужен какой-то другой тест для контроля орядока результатов
    # assert [result["ID"] for result in results] == ID_list, "Incorrect order of IDs"


def test_call_list_empty(bx_dummy):
    from tests.real_responses.batch_list_empty import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.call(
        "crm.lead.list", ({"filter": {"PHONE": "+0000877578564"}, "select": ["ID"]},)
    )
    assert isinstance(results, list)
    assert len(results) == 0

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.call(
        "crm.lead.list", {"filter": {"PHONE": "+0000877578564"}, "select": ["ID"]}
    )
    assert isinstance(results, list)
    assert len(results) == 0


def test_get_all_non_list_method(bx_dummy, recwarn):
    from tests.real_responses.user_fields import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_all("user.fields")
    assert isinstance(results, dict)
    print(w.message for w in recwarn.list)
    with pytest.raises(Exception):
        recwarn.pop(RuntimeWarning)


def test_batch_and_call_raw(bx_dummy):
    from tests.real_responses.call_batch import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.call_batch(
        {
            "halt": 0,
            "cmd": {
                "statuses": "crm.status.entity.items?entityId=STATUS",
                "sources": "crm.status.entity.items?entityId=SOURCE",
            },
        }
    )
    assert isinstance(results, dict)
    assert results.keys() == {"statuses", "sources"}

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.call(
        "batch",
        {
            "halt": 0,
            "cmd": {
                "statuses": "crm.status.entity.items?entityId=STATUS",
                "sources": "crm.status.entity.items?entityId=SOURCE",
            },
        },
        raw=True,
    )
    assert isinstance(results, dict)
    assert results.keys() == {"result", "time"}


def test_stagehistory(bx_dummy, recwarn):
    from tests.real_responses.stagehistory import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_all(
        "crm.stagehistory.list",
        {
            "entityTypeId": 2,
            "filter": {"STAGE_ID": "UC_KW9988"},
        },
    )

    assert isinstance(results, list)
    print(w.message for w in recwarn.list)
    with pytest.raises(Exception):
        recwarn.pop(RuntimeWarning)


def test_single_add(bx_dummy):
    from tests.real_responses.single_add import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.call(
        "tasks.task.add",
        {
            "fields": {
                "TITLE": "АГ",
                "DESCRIPTION": "Сложно сказать, почему потускнели светлые лики икон",
                "RESPONSIBLE_ID": 1,
                "TAGS": "скпэ22-4",
                "DEADLINE": "2022/07/30",
            }
        },
    )
    assert isinstance(results, dict)


def test_get_all_tasks(bx_dummy):
    from tests.real_responses.get_all_tasks import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_all(
        "tasks.task.list",
        {"filter": {"status": -1}, "select": ["ID", "STATUS", "STATUS_COMPLETE"]},
    )
    assert isinstance(results, list)
    assert len(results) == 1570


def test_crm_dealcategory_stage_list(bx_dummy):
    from tests.real_responses.crm_dealcategory_stage_list import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_by_ID("crm.dealcategory.stage.list", ["2", "4", "8", 0])
    assert isinstance(results, dict)


@pytest.mark.skip(
    "Абсолютно дикий формат ответа от Битрикс, нарушающий все шаблоны, "
    "которые присутствуют в других кейсах. А именно: батчевый ответ должен "
    "содержать метки переданных в батче команд и иметь структуру словаря списков. "
    "Тут - список списков."
)
def test_crm_dealcategory_stage_list_batch_of_one(bx_dummy):
    from tests.real_responses.crm_dealcategory_stage_list_batch_of_one import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_by_ID("crm.dealcategory.stage.list", [0])

    # should return a list of dicts
    assert isinstance(results, list) and isinstance(results[0], dict)


def test_catalog_document_element_list(bx_dummy):
    from tests.real_responses.catalog_document_element_list import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_all("catalog.document.element.list")

    assert len(results) == 95


def test_crm_item_productrow_list(bx_dummy):
    from tests.real_responses.crm_item_productrow_list import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_all("crm.item.productrow.list")

    assert len(results) == 2


def test_crm_stagehistory_list(bx_dummy):
    from tests.real_responses.crm_stagehistory_list import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_all("crm.stagehistory.list")

    assert len(results) == 9


def test_socialnetwork_api_workgroup_list(bx_dummy):
    from tests.real_responses.socialnetwork_api_workgroup_list import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_all(
        "socialnetwork.api.workgroup.list",
        {
            "select": [
                "ID",
                "NAME",
                "DESCRIPTION",
                "DATE_CREATE",
                "DATE_UPDATE",
                "DATE_ACTIVITY",
                "SUBJECT_ID",
                "KEYWORDS",
                "IMAGE_ID",
                "NUMBER_OF_MEMBERS",
                "INITIATE_PERMS",
                "SPAM_PERMS",
                "SUBJECT_NAME",
            ]
        },
    )

    assert len(results) == 50


def test_crm_add_item(bx_dummy):
    from tests.real_responses.crm_item_add import response

    bx_dummy.srh = MockSRH(response)
    result = bx_dummy.call(
        "crm.item.add", {"entityTypeId": 1, "fields": {"first_name": "first_name"}}
    )
    assert result != "item" and isinstance(result, dict) and "id" in result


def test_crm_item_list(bx_dummy):
    from tests.real_responses.crm_item_list import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_all("crm.item.list")

    assert len(results) == 3


def test_crm_category_list(bx_dummy):
    from tests.real_responses.crm_category_list import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_all("crm.category.list", {"entityTypeId": 2})

    assert results is not None
    assert isinstance(results, list)
    assert len(results) > 0
    assert all(isinstance(cat, dict) for cat in results)
    assert results == response["result"]["categories"]


def test_crm_contact_add_batch(bx_dummy):
    from tests.real_responses.crm_contact_add_batch import response

    bx_dummy.srh = MockSRH(response)
    result = bx_dummy.call(
        "crm.contact.add",
        {
            "fields": {
                "NAME": "TESTR",
                "PHONE": [{"VALUE": "78966666647", "VALUE_TYPE": "WORK"}],
                "ASSIGNED_BY_ID": -1,
                "OPENED": "Y",
            },
            "params": {"REGISTER_SONET_EVENT": "Y"},
        },
    )

    # Проверяем что результат - это ID созданного контакта
    assert result == 58943


def test_crm_contact_list(bx_dummy):
    response = {
        "result": [
            {
                "ID": "10",
                "NAME": "Абдуалимова Татьяна Александровна",
                "SECOND_NAME": None,
                "LAST_NAME": None,
            }
        ],
        "total": 1,
        "time": {
            "start": 1731743829.0188,
            "finish": 1731743829.06444,
            "duration": 0.045639991760253906,
            "processing": 0.019975900650024414,
            "date_start": "2024-11-16T10:57:09+03:00",
            "date_finish": "2024-11-16T10:57:09+03:00",
            "operating_reset_at": 1731744429,
            "operating": 0,
        },
    }

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_all(
        "crm.contact.list",
        {
            "params": {
                "select": ["ID", "NAME", "SECOND_NAME", "LAST_NAME"],
                "filter": {"=ID": ["10"]},
            }
        },
    )

    assert isinstance(results, list)
    assert len(results) == 1
    assert results[0]["ID"] == "10"
    assert results[0]["NAME"] == "Абдуалимова Татьяна Александровна"


def test_crm_company_contact_items_get(bx_dummy):
    from tests.real_responses.crm_company_contact_items_get import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_by_ID("crm.company.contact.items.get", ["205364"])
    assert len(results) == 2


def test_task_elapseditem_getlist_mixed_types(bx_dummy):
    """Тест для проверки обработки смешанных типов в get_by_ID для task.elapseditem.getlist.

    Проверяет, что get_by_ID корректно обрабатывает случаи,
    когда значения могут быть пустым списком, списком словарей или одиночным словарём.
    """
    from tests.real_responses.task_elapseditem_getlist_mixed_types import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.get_by_ID(
        "task.elapseditem.getlist",
        ["145572", "144518", "145620", "145485", "145359", "145523", "145142", "145476", "132990"]
    )

    assert isinstance(results, dict)

    # Пустой список
    assert "145572" in results
    assert isinstance(results["145572"], list)
    assert len(results["145572"]) == 0

    # Список из двух словарей
    assert "144518" in results
    assert isinstance(results["144518"], list)
    assert len(results["144518"]) == 2
    assert all(isinstance(item, dict) for item in results["144518"])
    assert results["144518"][0]["ID"] == "111301"
    assert results["144518"][1]["ID"] == "112004"

    # Список из одного словаря
    assert "145620" in results
    assert isinstance(results["145620"], list)
    assert len(results["145620"]) == 1
    assert results["145620"][0]["ID"] == "112248"

    # Одиночный словарь (в реальном ответе такого нет, но если появится)
    # assert isinstance(results["145485"], dict)  # В реальном ответе теперь это пустой список
    assert "145485" in results
    assert isinstance(results["145485"], list)
    assert len(results["145485"]) == 0

    # Список из одного словаря
    assert "145359" in results
    assert isinstance(results["145359"], list)
    assert len(results["145359"]) == 1
    assert results["145359"][0]["ID"] == "112058"

    # Список из одного словаря
    assert "145523" in results
    assert isinstance(results["145523"], list)
    assert len(results["145523"]) == 1
    assert results["145523"][0]["ID"] == "112249"

    # Список из одного словаря
    assert "145142" in results
    assert isinstance(results["145142"], list)
    assert len(results["145142"]) == 1
    assert results["145142"][0]["ID"] == "112244"

    # Список из нескольких словарей
    assert "145476" in results
    assert isinstance(results["145476"], list)
    assert len(results["145476"]) == 6
    assert results["145476"][0]["ID"] == "112063"
    assert results["145476"][-1]["ID"] == "112226"

    # Список из одного словаря
    assert "132990" in results
    assert isinstance(results["132990"], list)
    assert len(results["132990"]) == 1
    assert results["132990"][0]["ID"] == "112237"


def test_lists_element_add_batch(bx_dummy):
    """Тест для проверки правильного возвращения списка простых значений в batch ответе.

    Проверяет, что при batch-вызове lists.element.add возвращается список всех ID,
    а не только первый ID.
    """
    from tests.real_responses.lists_element_add_batch import response

    bx_dummy.srh = MockSRH(response)
    results = bx_dummy.call(
        "lists.element.add",
        [
            {
                "IBLOCK_TYPE_ID": "lists",
                "IBLOCK_ID": 33,
                "ELEMENT_CODE": "1",
                "FIELDS": {"NAME": "Тест 1", "PROPERTY_117": "1"}
            },
            {
                "IBLOCK_TYPE_ID": "lists",
                "IBLOCK_ID": 33,
                "ELEMENT_CODE": "2",
                "FIELDS": {"NAME": "Тест 2", "PROPERTY_117": "2"}
            }
        ]
    )

    # Ожидаем получить кортеж из двух значений (53, 55)
    assert isinstance(results, tuple)
    assert len(results) == 2
    assert results == (53, 55)



import contextlib
import pytest
from unittest.mock import AsyncMock, Mock
from fast_bitrix24.srh import ServerRequestHandler
import aiohttp


@pytest.mark.asyncio
async def test_request_attempt():
    # Create a mock response
    mock_response = Mock(spec=aiohttp.ClientResponse)
    mock_response.status = 200
    mock_response.json.return_value = {"time": {"operating": 1000}}

    @contextlib.asynccontextmanager
    async def mock_post(url, json, ssl):
        yield mock_response

    mock_session = AsyncMock()
    mock_session.post = mock_post

    handler = ServerRequestHandler(
        "https://google.com/webhook", None, True, 50, 2, 480, mock_session
    )

    # Call the method
    result = await handler.request_attempt("method", {"param": "value"})

    # Assert the expected behavior
    assert result == {"time": {"operating": 1000}}
    assert "method" in handler.method_throttlers



import warnings
from unittest.mock import MagicMock

import pytest

from fast_bitrix24.utils import get_warning_stack_level


async def empty_async(*args, **kwargs):
    pass


@pytest.mark.asyncio
async def test_warning_get_all(bx_dummy_async, monkeypatch):
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        monkeypatch.setattr(
            "fast_bitrix24.user_request.GetAllUserRequest.run",
            lambda *args, **kwargs: empty_async(),
        )
        await bx_dummy_async.get_all("crm.deal.add")
    assert len(w) == 1
    print(w[0])
    assert w[0].filename.lower() == __file__.lower()  # Assuming the test is in the same file


@pytest.mark.asyncio
async def test_warning_get_all_params(bx_dummy_async, monkeypatch):
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        monkeypatch.setattr(
            "fast_bitrix24.user_request.GetAllUserRequest.run",
            lambda *args, **kwargs: empty_async(),
        )
        await bx_dummy_async.get_all("crm.deal.list", params={"filter": {"ID": None}})
    assert len(w) == 1
    print(w[0])
    assert w[0].filename.lower() == __file__.lower()  # Assuming the test is in the same file



def mock_stack(filenames):
    frames = []
    for filename in filenames:
        frame = MagicMock()
        frame.f_back = frames[-1] if frames else None
        frame.f_code.co_filename = filename
        frames.append(frame)
    return frames


def mock_get_frame(stack, depth):
    return stack[-1 - depth]


# Test cases for happy path
@pytest.mark.parametrize(
    "module_filenames, module_sequence, expected_stack_level, test_id",
    [
        (["utils.py"], ["base.py", "utils.py", "caller.py"], 2, "single_module"),
        (
            ["module2.py", "module1.py"],
            ["base.py", "module1.py", "module2.py", "caller.py"],
            3,
            "multiple_modules",
        ),
        (
            ["module.py"],
            ["base.py", "module.py", "intermediate.py", "caller.py"],
            3,
            "intermediate_module",
        ),
        (
            "module.py",
            ["base.py", "module.py", "caller.py"],
            2,
            "single_module_str_input",
        ),
    ],
)
def test_get_warning_stack_level_happy_path(
    module_filenames, module_sequence, expected_stack_level, test_id, monkeypatch
):
    # Arrange
    monkeypatch.setattr(
        "fast_bitrix24.utils.sys._getframe",
        lambda depth=0: mock_get_frame(mock_stack(module_sequence), depth),
    )

    # Act
    stack_level = get_warning_stack_level(module_filenames)

    # Assert
    assert stack_level == expected_stack_level


# Test cases for error cases
@pytest.mark.parametrize(
    "module_filenames, module_sequence, expected_exception, test_id",
    [
        (["utils.py"], ["unrelated.py", "caller.py"], ValueError, "unrelated_module"),
    ],
)
def test_get_warning_stack_level_error_cases(
    module_filenames, module_sequence, expected_exception, test_id, monkeypatch
):
    # Arrange
    monkeypatch.setattr(
        "fast_bitrix24.utils.sys._getframe",
        lambda depth=0: mock_get_frame(mock_stack(module_sequence), depth),
    )

    # Act & Assert
    with pytest.raises(expected_exception):
        get_warning_stack_level(module_filenames)



from logging import DEBUG, NullHandler, getLogger
from .__version__ import __version__

logger = getLogger("fast_bitrix24")
logger.setLevel(DEBUG)
logger.addHandler(NullHandler())

def log(func):
    async def wrapper(*args, **kwargs):
        logger.info(f"Starting {func.__name__}({args}, {kwargs})")
        logger.debug(f"fast_bitrix24 version: {__version__}")
        try:
            from IPython import get_ipython

            logger.debug(f"IPython: {get_ipython()}")
        except ImportError:
            logger.debug("No IPython found")

        return await func(*args, **kwargs)

    return wrapper



from asyncio import Event, TimeoutError, sleep
from contextlib import asynccontextmanager

import aiohttp
from aiohttp.client_exceptions import (
    ClientConnectionError,
    ClientPayloadError,
    ClientResponseError,
)

from .throttle import SlidingWindowThrottler, LeakyBucketThrottler
from .logger import logger
from .utils import _url_valid

BITRIX_MAX_CONCURRENT_REQUESTS = 50

BITRIX_MEASUREMENT_PERIOD = 10 * 60

MAX_RETRIES = 10

RESTORE_CONNECTIONS_FACTOR = 1.3  # скорость восстановления количества запросов
DECREASE_CONNECTIONS_FACTOR = 3  # скорость уменьшения количества запросов
INITIAL_TIMEOUT = 0.5  # начальный таймаут в секундах
BACKOFF_FACTOR = 1.5  # основа расчета таймаута
# количество ошибок, до достижения котрого таймауты не делаются
NUM_FAILURES_NO_TIMEOUT = 3


class ServerError(Exception):
    pass


class TokenRejectedError(Exception):
    pass


RETRIED_ERRORS = (
    ClientPayloadError,
    ClientConnectionError,
    ServerError,
    TimeoutError,
)


class ServerRequestHandler:
    """
    Используется для контроля скорости доступа к серверам Битрикс.

    Основная цель - вести учет количества запросов, которые можно передать
    серверу Битрикс без получения ошибки `5XX`.
    """

    def __init__(
        self,
        webhook: str,
        token_func,
        respect_velocity_policy: bool,
        request_pool_size: int,
        requests_per_second: float,
        operating_time_limit: int,
        client,
        ssl: bool = True,
    ):
        self.webhook = self.standardize_webhook(webhook)

        self.token_func = token_func
        self.token = None

        # token_received - флаг, что получение токена начало и не закончено
        self.token_received = Event()
        self.token_received.set()

        self.respect_velocity_policy = respect_velocity_policy

        self.active_runs = 0

        self.operating_time_limit = operating_time_limit

        # если пользователь при инициализации передал клиента со своими настройками,
        # то будем использовать его клиента
        self.client_provided_by_user = bool(client)
        self.session = client
        self.ssl = ssl

        # лимит количества одновременных запросов,
        # установленный конструктором или пользователем
        self.mcr_max = BITRIX_MAX_CONCURRENT_REQUESTS

        # временный лимит количества одновременных запросов,
        # установленный через autothrottling
        self.mcr_cur_limit = BITRIX_MAX_CONCURRENT_REQUESTS

        self.concurrent_requests = 0
        self.request_complete = Event()

        # если положительное - количество последовательных удачных запросов
        # если отрицательное - количество последовательно полученных ошибок
        self.successive_results = 0

        # rate throttlers by method
        self.method_throttlers = {}  # dict[str, LeakyBucketLimiter]

        self.leaky_bucket_throttler = LeakyBucketThrottler(
            request_pool_size, requests_per_second
        )

    @staticmethod
    def standardize_webhook(webhook):
        """Приводит `webhook` к стандартному виду."""

        if not isinstance(webhook, str):
            raise TypeError(f"Webhook should be a {str}")

        webhook = webhook.strip()

        if not _url_valid(webhook):
            raise ValueError("Webhook is not a valid URL")

        if webhook[-1] != "/":
            webhook += "/"

        return webhook

    async def run_async(self, coroutine):
        """Запускает `coroutine`, создавая и прекращая сессию
        при необходимости."""

        async with self.handle_sessions():
            return await coroutine

    @asynccontextmanager
    async def handle_sessions(self):
        """Открывает и закрывает сессию в зависимости от наличия
        активных запросов."""

        # если клиент был задан пользователем, то ожидаем,
        # что пользователь сам откроет и закроет сессию

        if self.client_provided_by_user:
            yield
            return

        if not self.active_runs and (not self.session or self.session.closed):
            self.session = aiohttp.ClientSession(raise_for_status=True)
        self.active_runs += 1

        try:
            yield

        finally:
            self.active_runs -= 1
            if not self.active_runs and self.session and not self.session.closed:
                await self.session.close()

    async def single_request(self, method: str, params=None) -> dict:
        """Делает единичный запрос к серверу,
        с повторными попытками при необходимости."""

        # начальное получение токена
        if self.token_func and not self.token:
            await self.ensure_new_token()

        while True:

            try:
                result = await self.request_attempt(method.strip().lower(), params)
                self.success()
                return result

            except TokenRejectedError:
                await self.ensure_new_token()

            except RETRIED_ERRORS as err:
                self.failure(err)

            # all other exceptions will propagate

    async def request_attempt(self, method, params=None) -> dict:
        """Делает попытку запроса к серверу, ожидая при необходимости."""

        try:
            async with self.acquire(method):
                logger.debug(f"Requesting {{'method': {method}, 'params': {params}}}")

                params_with_auth = params.copy() if params else {}
                if self.token:
                    params_with_auth["auth"] = self.token

                async with self.session.post(
                    url=self.webhook + method, json=params_with_auth, ssl=self.ssl
                ) as response:
                    json = await response.json(encoding="utf-8")

                    logger.debug("Response: %s", json)

                    self.add_throttler_records(method, params, json)

                    return json

        except ClientResponseError as error:

            if error.status // 100 == 5:  # ошибки вида 5XX
                raise ServerError("The server returned an error") from error

            elif error.status == 401 and self.token_func:
                raise TokenRejectedError(
                    "The server rejected the auth token"
                ) from error

            raise  # иначе повторяем полученное исключение

    def add_throttler_records(self, method, params: dict, json: dict):
        if "result_time" in json:
            for cmd_name, cmd_url in params["cmd"].items():
                item_method = cmd_url.split("?")[0]
                item_time = json["result_time"][cmd_name]
                self.method_throttlers[item_method].add_request_record(item_time)
        elif "operating" in json["time"]:
            request_run_time = json["time"]["operating"]
            self.method_throttlers[method].add_request_record(request_run_time)

        self.leaky_bucket_throttler.add_request_record()

    def success(self):
        """Увеличить счетчик удачных попыток."""

        self.successive_results = max(self.successive_results + 1, 1)

    def failure(self, err: Exception):
        """Увеличить счетчик неудачных попыток и поднять исключение,
        если попытки исчерпаны."""

        self.successive_results = min(self.successive_results - 1, -1)

        if self.successive_results < -MAX_RETRIES:
            raise RuntimeError(
                "All attempts to get data from server exhausted"
            ) from err

    @asynccontextmanager
    async def acquire(self, method: str):
        """Ожидает, пока не станет безопасно делать запрос к серверу."""

        await self.autothrottle()

        async with self.limit_concurrent_requests(), self.leaky_bucket_throttler.acquire():
            if self.respect_velocity_policy:
                if method not in self.method_throttlers:
                    self.method_throttlers[method] = SlidingWindowThrottler(
                        self.operating_time_limit, BITRIX_MEASUREMENT_PERIOD
                    )

                async with self.method_throttlers[method].acquire():
                    yield

            else:
                yield

    async def autothrottle(self):
        """Если было несколько неудач, делаем таймаут и уменьшаем скорость
        и количество одновременных запросов, и наоборот."""

        if self.successive_results < 0:
            self.mcr_cur_limit = max(
                self.mcr_cur_limit / DECREASE_CONNECTIONS_FACTOR, 1
            )

            logger.debug(
                f"Concurrent requests decreased: {{'mcr_cur_limit': {self.mcr_cur_limit}}}"
            )

            if self.successive_results < NUM_FAILURES_NO_TIMEOUT:
                power = -self.successive_results - NUM_FAILURES_NO_TIMEOUT - 1
                delay = INITIAL_TIMEOUT * BACKOFF_FACTOR**power

                logger.debug(f"Delaying request: {{'delay': {delay}}}")

                await sleep(delay)

        elif self.successive_results > 0:

            self.mcr_cur_limit = min(
                self.mcr_cur_limit * RESTORE_CONNECTIONS_FACTOR, self.mcr_max
            )

            logger.debug(
                f"Concurrent requests increased: {{'mcr_cur_limit': {self.mcr_cur_limit}}}"
            )

    @asynccontextmanager
    async def limit_concurrent_requests(self):
        """Не позволяет одновременно выполнять
        более `self.mcr_cur_limit` запросов."""

        while self.concurrent_requests > self.mcr_cur_limit:
            self.request_complete.clear()
            await self.request_complete.wait()

        self.concurrent_requests += 1

        try:
            yield

        finally:
            self.concurrent_requests -= 1
            self.request_complete.set()

    async def ensure_new_token(self):
        """Получает новый токен, если процесс получения токена еще не запущен,
        или ждет его завершения."""

        if self.token_received.is_set():
            logger.debug("Requesting new token")

            self.token_received.clear()
            self.token = await self.token_func()
            self.token_received.set()
        else:
            await self.token_received.wait()



import sys
from typing import List, Union
from urllib.parse import quote, urlparse


def _url_valid(url):
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc, result.path])
    except Exception:
        return False


def http_build_query(params, convention="%s"):
    if len(params) == 0:
        return ""

    output = ""
    for key in params.keys():

        if type(params[key]) is dict:

            output += http_build_query(params[key], convention % key + "[%s]")

        elif type(params[key]) is list:

            new_params = {str(i): element for i, element in enumerate(params[key])}

            output += http_build_query(new_params, convention % key + "[%s]")

        else:

            val = quote(str(params[key]))
            key = quote(key)
            output = output + convention % key + "=" + val + "&"

    return output


def get_warning_stack_level(module_filenames: Union[str, List[str]]) -> int:
    """Calculate the stack level for warnings issued from a library.

    Returns the number of stack frames between the top-most module from the given list
    occurring in the call stack and the caller of the function.

    This is used to provide the appropriate stack level to warnings.warn or
    warnings.warn_explicit so that the warning appears in user code rather
    than in the library itself.

    The calculation is based on the system call stack.

    Args:
        module_filenames: The filenames of possible top-most modules before which the
            user code is situated. If a single filename is provided, it is wrapped
            in a list.

    Returns:
        The stack level to be used in warnings.warn or warnings.warn_explicit.

    Raises:
        ValueError: If none of the modules in the provided sequence have been found
            in the stack.
    """

    # Wrap single filename in a list
    if isinstance(module_filenames, str):
        module_filenames = [module_filenames]

    # Append '.py' to filenames
    module_filenames = tuple(
        f if f.endswith(".py") else f"{f}.py" for f in module_filenames
    )

    # Get filenames from stack
    stack_filenames = []
    top_frame = sys._getframe()
    while top_frame is not None:
        stack_filenames.append(top_frame.f_code.co_filename)
        top_frame = top_frame.f_back

    # Find top-most module filename in the stack
    try:
        top_most_index = next(
            i
            for i, v in enumerate(reversed(stack_filenames))
            if v.endswith(module_filenames)
        )
    except StopIteration:
        raise ValueError(
            "None of the modules in the provided sequence have been found "
            "in the stack."
        )

    # Calculate stack level
    return len(stack_filenames) - top_most_index



import os
from asyncio import create_task, ensure_future, gather, sleep, wait
from collections import namedtuple
from contextlib import asynccontextmanager
from time import monotonic

import pytest
from fast_bitrix24 import BitrixAsync
from fast_bitrix24.srh import ServerRequestHandler


@pytest.mark.skipif(
    not os.getenv("FAST_BITRIX24_TEST_WEBHOOK"),
    reason="Нет аккаунта, на котором можно проверить",
)
@pytest.mark.asyncio
class TestAsync:
    async def test_simple_async_calls(self, create_100_leads_async):

        b: BitrixAsync = create_100_leads_async

        leads = await b.get_all("crm.lead.list")
        await b.get_by_ID("crm.lead.get", [lead["ID"] for lead in leads])
        await b.list_and_get("crm.lead")
        await b.call("crm.lead.get", {"ID": leads[0]["ID"]})
        await b.call_batch({"halt": 0, "cmd": {0: "crm.lead.list"}})

    async def test_simultaneous_calls(self, create_100_leads_async):
        b = create_100_leads_async

        result = await gather(
            b.get_all("crm.lead.list"),
            b.get_all("crm.lead.list"),
            b.get_all("crm.lead.list"),
        )

        assert len(result) == 3
        assert result[0] == result[1] == result[2]
        assert all(len(r) >= 100 for r in result)
